# flake8: noqa
from .auth_views import login_for_access_token
from .emulation_views import emulation_run_list_handler
from .extra_views import docs, favicon
from .feature_views import get_features_handler
from .tags_views import tags_item_handler, tags_list_handler
from .testrun_views import get_test_run_handler, run_tests_by_tag_handler
from .testuser_views import (
    delete_test_user_handler,
    get_test_user_handler,
    test_user_get_spec_handler,
    test_user_list_handler,
    test_user_put_spec_handler,
)



import abc
from typing import Optional, cast

from pydantic import SecretStr

from overhave import db
from overhave.storage.converters import SystemUserModel


class BaseSystemStorageException(Exception):
    """Base exception for :class:`SystemUserStorage`."""


class SystemUserNotFoundError(Exception):
    """Error for situation without user with given id."""


class ISystemUserStorage(abc.ABC):
    """Abstract class for system user storage."""

    @staticmethod
    @abc.abstractmethod
    def get_user(user_id: int) -> Optional[SystemUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_user(login: str, password: Optional[SecretStr] = None, role: db.Role = db.Role.user) -> SystemUserModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_user_by_credits(login: str, password: Optional[SecretStr] = None) -> Optional[SystemUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_user_role(user_model: SystemUserModel) -> None:
        pass


class SystemUserStorage(ISystemUserStorage):
    """Class for system user storage."""

    @staticmethod
    def get_user(user_id: int) -> Optional[SystemUserModel]:
        with db.create_session() as session:
            db_user = session.query(db.UserRole).get(user_id)
            if db_user is not None:
                return cast(SystemUserModel, SystemUserModel.from_orm(db_user))
            return None

    @staticmethod
    def create_user(login: str, password: Optional[SecretStr] = None, role: db.Role = db.Role.user) -> SystemUserModel:
        with db.create_session() as session:
            db_password = None
            if password is not None:
                db_password = password.get_secret_value()
            db_user = db.UserRole(login=login, password=db_password, role=role)
            session.add(db_user)
            session.flush()
            return cast(SystemUserModel, SystemUserModel.from_orm(db_user))

    @staticmethod
    def get_user_by_credits(login: str, password: Optional[SecretStr] = None) -> Optional[SystemUserModel]:
        with db.create_session() as session:
            query = session.query(db.UserRole).filter(db.UserRole.login == login)
            if password is not None:
                query = query.filter(db.UserRole.password == password.get_secret_value())
            db_user = query.one_or_none()
            if db_user is not None:
                return cast(SystemUserModel, SystemUserModel.from_orm(db_user))
            return None

    @staticmethod
    def update_user_role(user_model: SystemUserModel) -> None:
        with db.create_session() as session:
            db_user = session.query(db.UserRole).get(user_model.id)
            if db_user is None:
                raise SystemUserNotFoundError(f"User with id {user_model.id} was not found!")
            db_user.role = user_model.role



# flake8: noqa
from .compiler import ScenarioCompiler, generate_task_info
from .errors import IncorrectScenarioTextError
from .file_manager import FileManager
from .parser import FeatureInfo, ScenarioParser



from datetime import timedelta

from pydantic.env_settings import BaseSettings
from pydantic.types import SecretStr

from overhave.base_settings import OVERHAVE_ENV_PREFIX


class OverhaveUvicornApiSettings(BaseSettings):
    """Settings for Overhave API server, started with Uvicorn."""

    host: str = "0.0.0.0"
    port: int = 8000
    workers: int = 1

    class Config:
        env_prefix = OVERHAVE_ENV_PREFIX + "UVICORN_"


class OverhaveApiAuthSettings(BaseSettings):
    """Settings for Overhave API service auth_managers."""

    secret_key: SecretStr
    algorithm: str = "HS256"
    access_token_expire_minutes: int = 30

    @property
    def access_token_expire_timedelta(self) -> timedelta:
        return timedelta(minutes=self.access_token_expire_minutes)

    class Config:
        env_prefix = OVERHAVE_ENV_PREFIX + "API_AUTH_"



from typing import List

import fastapi

from overhave.api.auth import get_authorized_user
from overhave.api.views import (
    delete_test_user_handler,
    docs,
    emulation_run_list_handler,
    favicon,
    get_features_handler,
    get_test_run_handler,
    get_test_user_handler,
    login_for_access_token,
    run_tests_by_tag_handler,
    tags_item_handler,
    tags_list_handler,
    test_user_get_spec_handler,
    test_user_list_handler,
    test_user_put_spec_handler,
)
from overhave.storage import (
    AuthToken,
    EmulationRunModel,
    FeatureModel,
    TagModel,
    TestRunModel,
    TestUserModel,
    TestUserSpecification,
)


def _get_tags_router() -> fastapi.APIRouter:
    tags_router = fastapi.APIRouter()
    tags_router.add_api_route(
        "/item",
        tags_item_handler,
        methods=["GET"],
        response_model=TagModel,
        summary="Get FeatureTag",
        description="Get feature tag by `value`",
    )
    tags_router.add_api_route(
        "/list",
        tags_list_handler,
        response_model=List[TagModel],
        methods=["GET"],
        summary="Get FeatureTags",
        description="Get list of feature tags like `value`",
    )
    return tags_router


def _get_feature_router() -> fastapi.APIRouter:
    feature_router = fastapi.APIRouter()
    feature_router.add_api_route(
        "/",
        get_features_handler,
        methods=["GET"],
        response_model=List[FeatureModel],
        summary="Get list of Feature info",
        description="Get list of feature info by `tag_id` or `tag_value`",
    )
    return feature_router


def _get_testrun_router() -> fastapi.APIRouter:
    test_run_router = fastapi.APIRouter()
    test_run_router.add_api_route(
        "/",
        get_test_run_handler,
        methods=["GET"],
        response_model=TestRunModel,
        summary="Get test run by id",
        description="Get test run by `id`",
    )
    test_run_router.add_api_route(
        "/create/",
        run_tests_by_tag_handler,
        methods=["POST"],
        response_model=list[str],
        summary="Create TestRunTasks for Features by tag_value",
        description="Create TestRunTasks for Features by `tag_value`",
    )
    return test_run_router


def _get_testuser_router() -> fastapi.APIRouter:
    test_user_router = fastapi.APIRouter()

    test_user_router.add_api_route(
        "/",
        get_test_user_handler,
        methods=["GET"],
        response_model=TestUserModel,
        summary="Get TestUser",
        description="Get test user full info by `user_id` or `user_name`",
    )
    test_user_router.add_api_route(
        "/list",
        test_user_list_handler,
        methods=["GET"],
        response_model=List[TestUserModel],
        summary="Get list of TestUsers",
        description="Get list of test users with given `feature_type` and `allow_update`",
    )
    test_user_router.add_api_route(
        "/{user_id}/delete",
        delete_test_user_handler,
        methods=["DELETE"],
        summary="Delete TestUser",
        description="Delete test user by `user_id`",
    )
    test_user_router.add_api_route(
        "/{user_id}/specification",
        test_user_get_spec_handler,
        methods=["GET"],
        response_model=TestUserSpecification,
        summary="Get TestUser specification",
        description="Get test user specification by `user_id`",
    )
    test_user_router.add_api_route(
        "/{user_id}/specification",
        test_user_put_spec_handler,
        methods=["PUT"],
        summary="Put test user specification",
        description="Update test user specification by `user_id` and given payload",
    )
    return test_user_router


def _get_emulation_router() -> fastapi.APIRouter:
    emulation_router = fastapi.APIRouter()
    emulation_router.add_api_route(
        "/run/list",
        emulation_run_list_handler,
        methods=["GET"],
        response_model=List[EmulationRunModel],
        summary="Get list of EmulationRun info",
        description="Get list of EmulationRun info by `test_user_id`",
    )
    return emulation_router


def _get_auth_router() -> fastapi.APIRouter:
    auth_router = fastapi.APIRouter()
    auth_router.add_api_route(
        "/token",
        login_for_access_token,
        methods=["POST"],
        response_model=AuthToken,
        summary="Get OAuth2 token",
        description="Get OAuth2 token by username and password",
        include_in_schema=False,
    )
    return auth_router


def create_overhave_api() -> fastapi.FastAPI:
    app = fastapi.FastAPI()
    auth_deps = [fastapi.Depends(get_authorized_user)]

    app.include_router(_get_tags_router(), dependencies=auth_deps, prefix="/feature/tags", tags=["feature_tags"])
    app.include_router(_get_feature_router(), dependencies=auth_deps, prefix="/feature", tags=["features"])
    app.include_router(_get_testuser_router(), dependencies=auth_deps, prefix="/test_user", tags=["test_users"])
    app.include_router(_get_testrun_router(), dependencies=auth_deps, prefix="/test_run", tags=["test_runs"])
    app.include_router(_get_emulation_router(), dependencies=auth_deps, prefix="/emulation", tags=["emulations"])

    app.include_router(_get_auth_router())
    app.add_api_route("/", docs, methods=["GET"], include_in_schema=False)
    app.add_api_route("/favicon.ico", favicon, methods=["GET"], include_in_schema=False)
    return app



# flake8: noqa
from .gitlab_publisher import GitlabVersionPublisher
from .settings import OverhaveGitlabPublisherSettings
from .tokenizer.client import TokenizerClient
from .tokenizer.settings import TokenizerClientSettings



import enum
from typing import Optional

import sqlalchemy as sa
import sqlalchemy_utils as su

from overhave.db.base import BaseTable, PrimaryKeyMixin


class Role(str, enum.Enum):
    """Enum that declares user access roles."""

    user = "user"
    admin = "admin"


@su.generic_repr("login", "role")
class UserRole(BaseTable, PrimaryKeyMixin):
    """User access table."""

    login = sa.Column(sa.String(), nullable=False, unique=True)
    password = sa.Column(sa.String(), nullable=True)
    role = sa.Column(sa.Enum(Role), nullable=False, default=Role.user)

    def __init__(self, login: str, password: Optional[str], role: Role) -> None:
        self.login = login
        self.password = password
        self.role = role

    def __repr__(self) -> str:
        return f"{self.login} ({self.role})"


@su.generic_repr("group")
class GroupRole(BaseTable, PrimaryKeyMixin):
    """Group access table."""

    group = sa.Column(sa.String(), nullable=False, unique=True)

    def __init__(self, group: str) -> None:
        self.group = group



# flake8: noqa
from .allure_utils import (
    DescriptionManager,
    StepContextNotDefinedError,
    StepContextRunner,
    add_admin_feature_link_to_report,
    add_task_links_to_report,
    set_severity_level,
)
from .bdd_item import (
    add_scenario_title_to_report,
    get_full_step_name,
    get_scenario,
    is_pytest_bdd_item,
    set_git_project_url_if_necessary,
)
from .parsed_info import get_feature_info_from_item, set_feature_info_for_item



import re
from typing import NewType

from pydantic import BaseModel, validator
from pytest_bdd.types import STEP_TYPES

StepTypeName = NewType("StepTypeName", str)


class BddStepModel(BaseModel):
    """Model for pytest_bdd step in Overhave."""

    type: StepTypeName
    name: str
    doc: str

    @validator("type")
    def validate_type(cls, v: str) -> StepTypeName:
        if v not in STEP_TYPES:
            raise ValueError
        return StepTypeName(v)

    @property
    def html_doc(self) -> str:
        return re.sub(r"\n\s{4}", "\n", self.doc).strip("\n ")



# flake8: noqa
from .admin import admin
from .api import api
from .consumers import consumer
from .db_cmds import set_config_to_context
from .group import overhave
from .synchronization import synchronize



import abc
from typing import Optional

from pydantic import SecretStr

from overhave.storage import SystemUserModel


class IAdminAuthorizationManager(abc.ABC):
    """Abstract class for auth_managers manager."""

    @abc.abstractmethod
    def authorize_user(self, username: str, password: SecretStr) -> Optional[SystemUserModel]:
        pass



# flake8: noqa
from .description_manager import DescriptionManager
from .links import add_admin_feature_link_to_report, add_git_project_feature_link_to_report, add_task_links_to_report
from .severity import set_severity_level
from .step_context_runner import StepContextNotDefinedError, StepContextRunner



from datetime import timedelta
from typing import Optional, Tuple, TypeVar

from pydantic import BaseSettings, validator
from yarl import URL

from overhave.utils import make_url


class BaseHttpClientSettings(BaseSettings):
    """Base settings for :class:`BaseHttpClient`."""

    url: URL

    connect_timeout: timedelta = timedelta(seconds=1)
    read_timeout: timedelta = timedelta(seconds=5)

    @validator("url", pre=True)
    def make_url(cls, v: Optional[str]) -> Optional[URL]:
        return make_url(v)

    @property
    def timeout(self) -> Tuple[float, float]:
        return self.connect_timeout.total_seconds(), self.read_timeout.total_seconds()


HttpSettingsType = TypeVar("HttpSettingsType", bound=BaseHttpClientSettings)



from pydantic.fields import Field
from pydantic.main import BaseModel


class AuthTokenData(BaseModel):
    """Model for OAuth2 auth_managers token data."""

    username: str


class AuthHeaders(BaseModel):
    """Model for OAuth2 auth_managers HTTP headers."""

    Authorization: str = Field("Bearer", const=True)


AUTH_HEADERS = AuthHeaders()



import abc
import logging
from pathlib import Path
from typing import Dict, List, Optional, Set

from overhave.entities import BaseFileExtractor, OverhaveFileSettings

logger = logging.getLogger(__name__)


class IPluginResolver(abc.ABC):
    """Abstract class for custom pytest-bdd steps modules resolution."""

    @abc.abstractmethod
    def get_plugins(self, plugin_type: Optional[str] = None) -> List[str]:
        pass


class PluginResolver(BaseFileExtractor, IPluginResolver):
    """Class for custom pytest-bdd steps modules resolution."""

    def __init__(self, file_settings: OverhaveFileSettings):
        super().__init__(extenstion=".py")
        self._file_settings = file_settings
        self._plugins = self._resolve_plugins(directory=self._file_settings.steps_dir)
        logger.debug("Available Overhave pytest plugin modules: %s", self._plugins)

    def _resolve_plugins(self, directory: Path) -> Dict[str, List[Path]]:
        plugin_folders = [d for d in directory.iterdir() if self._check_dir_compliance(d)]
        plugin_dict = {folder.name: self._extract_recursively(folder) for folder in plugin_folders}

        plugin_modules = [m for m in directory.iterdir() if self._check_file_compliance(m)]
        for folder in plugin_dict:
            plugin_dict[folder].extend(plugin_modules)
        return plugin_dict

    def _format_path(self, path: Path) -> str:
        relative_path = path.relative_to(self._file_settings.work_dir).as_posix()
        return relative_path.replace("/", ".").rstrip(".py")

    def get_plugins(self, plugin_type: Optional[str] = None) -> List[str]:
        if isinstance(plugin_type, str):
            if self._plugins.get(plugin_type) is None:
                raise KeyError(f"Specified plugin type '{plugin_type}' does not exist!")
            plugins = [self._format_path(x) for x in self._plugins[plugin_type]]
            logger.debug("Return plugins by plugin_type='%s': %s", plugin_type, plugins)
            return plugins

        joined_modules: Set[str] = set()
        for key in self._plugins:
            joined_modules.update(set(self.get_plugins(key)))
        modules_list = list(joined_modules)
        logger.debug("Return all existing plugins: %s", modules_list)
        return modules_list



from pydantic.fields import Field
from pydantic.main import BaseModel


class TokenRequestData(BaseModel):
    """Model for OAuth2 request data."""

    grant_type: str = Field("password", const=True)
    username: str
    password: str



# flake8: noqa
from .abstract import IAdminAuthorizationManager
from .default import DefaultAdminAuthorizationManager
from .ldap import LDAPAdminAuthorizationManager, OverhaveLdapManagerSettings
from .simple import SimpleAdminAuthorizationManager



from typing import Any, Dict, Optional

from pydantic import BaseSettings, root_validator


class OverhaveS3ManagerSettings(BaseSettings):
    """Settings for S3Client."""

    enabled: bool = False

    url: Optional[str]
    region_name: Optional[str]
    access_key: Optional[str]
    secret_key: Optional[str]
    verify: bool = True

    autocreate_buckets: bool = False

    class Config:
        env_prefix = "OVERHAVE_S3_"

    @root_validator
    def validate_enabling(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        enabled = values.get("enabled")
        if enabled and not all(
            (
                isinstance(values.get("url"), str),
                isinstance(values.get("access_key"), str),
                isinstance(values.get("secret_key"), str),
            )
        ):
            raise ValueError("Url, access key and secret key should be specified!")
        return values



from functools import cache
from pathlib import Path

from overhave.api.settings import OverhaveApiAuthSettings
from overhave.entities import OverhaveEmulationSettings, OverhaveRedisSettings
from overhave.storage import (
    DraftStorage,
    EmulationStorage,
    FeatureStorage,
    FeatureTagStorage,
    FeatureTypeStorage,
    IDraftStorage,
    IEmulationStorage,
    IFeatureStorage,
    IFeatureTagStorage,
    IFeatureTypeStorage,
    ISystemUserStorage,
    ITestRunStorage,
    ITestUserStorage,
    SystemUserStorage,
    TestRunStorage,
    TestUserStorage,
)
from overhave.transport import EmulationTask, PublicationTask, RedisProducer, RedisStream, TestRunTask


@cache
def get_api_auth_settings() -> OverhaveApiAuthSettings:
    return OverhaveApiAuthSettings()


@cache
def get_admin_files_dir() -> Path:
    return Path(__file__).parent.parent / "admin" / "files"


@cache
def get_system_user_storage() -> ISystemUserStorage:
    return SystemUserStorage()


@cache
def get_feature_tag_storage() -> IFeatureTagStorage:
    return FeatureTagStorage()


@cache
def get_feature_type_storage() -> IFeatureTypeStorage:
    return FeatureTypeStorage()


@cache
def get_feature_storage() -> IFeatureStorage:
    return FeatureStorage()


@cache
def get_draft_storage() -> IDraftStorage:
    return DraftStorage()


@cache
def get_test_user_storage() -> ITestUserStorage:
    return TestUserStorage()


@cache
def get_test_run_storage() -> ITestRunStorage:
    return TestRunStorage()


@cache
def get_emulation_storage() -> IEmulationStorage:
    return EmulationStorage(OverhaveEmulationSettings())


@cache
def get_redis_producer() -> RedisProducer:
    return RedisProducer(
        settings=OverhaveRedisSettings(),
        mapping={
            TestRunTask: RedisStream.TEST,
            PublicationTask: RedisStream.PUBLICATION,
            EmulationTask: RedisStream.EMULATION,
        },
    )



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import Extra, Field, root_validator
from pydantic.main import BaseModel


class BucketModel(BaseModel):
    """Model for boto3 client Bucket."""

    name: str = Field(alias="Name")
    created_at: datetime = Field(alias="CreationDate")


class OwnerModel(BaseModel):
    """Model for boto3 client object owner."""

    name: str = Field(alias="DisplayName")
    owner_id: str = Field(alias="ID")


class ObjectModel(BaseModel):
    """Model for boto3 client object."""

    name: str = Field(alias="Key")
    modified_at: datetime = Field(alias="LastModified")
    etag: str = Field(alias="ETag")
    size: int = Field(alias="Size")
    storage_class: str = Field(alias="StorageClass")
    owner: OwnerModel = Field(alias="Owner")

    class Config:
        extra = Extra.allow


class BaseObjectToDeletionModel(BaseModel):
    """Base model for boto3 client object deletion result."""

    name: str = Field(alias="Key")
    etag: Optional[str] = Field(alias="VersionId")


class DeletedObjectModel(BaseObjectToDeletionModel):
    """Model for boto3 client deleted object."""

    marker: Optional[bool] = Field(alias="DeleteMarker")
    marker_id: Optional[bool] = Field(alias="DeleteMarkerVersionId")


class NotDeletedObjectModel(BaseObjectToDeletionModel):
    """Model for boto3 client not deleted object."""

    code: str = Field(alias="Code")
    message: str = Field(alias="Message")


class DeletionResultModel(BaseModel):
    """Model for boto3 client objects deletion result."""

    deleted: Optional[List[DeletedObjectModel]] = Field(alias="Deleted")
    errors: Optional[List[NotDeletedObjectModel]] = Field(alias="Errors")
    requester: Optional[str] = Field(alias="RequestCharged")

    @root_validator(pre=True)
    def validate_results(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        deleted = values.get("Deleted")
        errors = values.get("Errors")
        if deleted is None and errors is None:
            raise ValueError("At least one result field should be presented!")
        return values



from pathlib import Path
from typing import List


class BaseFileExtractor:
    """Base class for file extraction."""

    def __init__(self, extenstion: str) -> None:
        self._extenstion = extenstion

    @staticmethod
    def _check_dir_compliance(path: Path) -> bool:
        return all((path.is_dir(), not path.name.startswith("."), not path.name.startswith("_")))

    def _check_file_compliance(self, path: Path) -> bool:
        return all(
            (
                not path.is_dir(),
                path.suffix == self._extenstion,
                not path.name.startswith("."),
                not path.name.startswith("_"),
            )
        )

    def _extract_recursively(self, folder: Path) -> List[Path]:
        files = []
        for path in folder.iterdir():
            if self._check_dir_compliance(path):
                subdirs = self._extract_recursively(path)
                files.extend(subdirs)
                continue
            if not self._check_file_compliance(path):
                continue
            files.append(path)
        return files



from typing import Optional, cast

from pydantic import BaseSettings

from overhave.entities import (
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveReportManagerSettings,
    OverhaveScenarioCompilerSettings,
)
from overhave.test_execution import OverhaveProjectSettings, OverhaveTestSettings
from overhave.transport import OverhaveS3ManagerSettings


class BaseFactoryContextException(Exception):
    """Base exception for :class:`BaseFactoryContext`."""


class NotDefinedSettingsError(BaseFactoryContextException):
    """Exception for situation with not defined specific settings."""


class BaseFactoryContext:
    """Base class for factory context."""

    def __init__(
        self,
        compilation_settings: Optional[OverhaveScenarioCompilerSettings] = None,
        emulation_settings: Optional[OverhaveEmulationSettings] = None,
        file_settings: Optional[OverhaveFileSettings] = None,
        language_settings: Optional[OverhaveLanguageSettings] = None,
        project_settings: Optional[OverhaveProjectSettings] = None,
        report_manager_settings: Optional[OverhaveReportManagerSettings] = None,
        s3_manager_settings: Optional[OverhaveS3ManagerSettings] = None,
        test_settings: Optional[OverhaveTestSettings] = None,
    ):
        self._compilation_settings = compilation_settings
        self._emulation_settings = emulation_settings
        self._file_settings = file_settings
        self._language_settings = language_settings
        self._project_settings = project_settings
        self._report_manager_settings = report_manager_settings
        self._s3_manager_settings = s3_manager_settings
        self._test_settings = test_settings

    @staticmethod
    def _ensured_settings(settings: Optional[BaseSettings]) -> BaseSettings:
        if settings is None:
            raise NotDefinedSettingsError(type(settings).__name__)
        return settings

    @property
    def compilation_settings(self) -> OverhaveScenarioCompilerSettings:
        return cast(OverhaveScenarioCompilerSettings, self._ensured_settings(self._compilation_settings))

    @property
    def emulation_settings(self) -> OverhaveEmulationSettings:
        return cast(OverhaveEmulationSettings, self._ensured_settings(self._emulation_settings))

    @property
    def file_settings(self) -> OverhaveFileSettings:
        return cast(OverhaveFileSettings, self._ensured_settings(self._file_settings))

    @property
    def language_settings(self) -> OverhaveLanguageSettings:
        return cast(OverhaveLanguageSettings, self._ensured_settings(self._language_settings))

    @property
    def project_settings(self) -> OverhaveProjectSettings:
        return cast(OverhaveProjectSettings, self._ensured_settings(self._project_settings))

    @property
    def report_manager_settings(self) -> OverhaveReportManagerSettings:
        return cast(OverhaveReportManagerSettings, self._ensured_settings(self._report_manager_settings))

    @property
    def s3_manager_settings(self) -> OverhaveS3ManagerSettings:
        return cast(OverhaveS3ManagerSettings, self._ensured_settings(self._s3_manager_settings))

    @property
    def test_settings(self) -> OverhaveTestSettings:
        return cast(OverhaveTestSettings, self._ensured_settings(self._test_settings))



import abc
from functools import cached_property

from overhave.factory.base_factory import BaseOverhaveFactory, IOverhaveFactory
from overhave.factory.context import OverhaveSynchronizerContext
from overhave.synchronization import IOverhaveSynchronizer, OverhaveSynchronizer


class ISynchronizerFactory(IOverhaveFactory[OverhaveSynchronizerContext]):
    """Factory interface for Overhave feature synchronizer application."""

    @property
    @abc.abstractmethod
    def synchronizer(self) -> IOverhaveSynchronizer:
        pass


class SynchronizerFactory(BaseOverhaveFactory[OverhaveSynchronizerContext], ISynchronizerFactory):
    """Factory for Overhave feature synchronization application."""

    context_cls = OverhaveSynchronizerContext

    @cached_property
    def _synchronizer(self) -> OverhaveSynchronizer:
        return OverhaveSynchronizer(
            file_settings=self.context.file_settings,
            feature_storage=self._feature_storage,
            feature_type_storage=self._feature_type_storage,
            scenario_storage=self._scenario_storage,
            draft_storage=self._draft_storage,
            tag_storage=self._feature_tag_storage,
            scenario_parser=self._scenario_parser,
            feature_extractor=self._feature_extractor,
            system_user_storage=self._system_user_storage,
            git_initializer=self._git_initializer,
        )

    @property
    def synchronizer(self) -> IOverhaveSynchronizer:
        return self._synchronizer



import re
from datetime import datetime
from typing import Any, Dict, List, Optional

import allure
from flask_admin.contrib.sqla import ModelView
from markupsafe import Markup
from yarl import URL

from overhave import db
from overhave.admin.views.formatters.helpers import (
    get_button_class_by_status,
    get_feature_link_markup,
    get_report_index_link,
    get_testrun_details_link,
)
from overhave.admin.views.formatters.safe_formatter import safe_formatter
from overhave.db import TestReportStatus


@safe_formatter(
    type=datetime,
    supported_models=(
        db.UserRole,
        db.GroupRole,
        db.Feature,
        db.TestRun,
        db.Draft,
        db.TestUser,
        db.Emulation,
        db.EmulationRun,
        db.Tags,
    ),
)
def datetime_formatter(view: ModelView, context: Any, model: db.BaseTable, value: datetime) -> Markup:
    return Markup(value.strftime("%d-%m-%Y %H:%M:%S"))


@safe_formatter(type=list, supported_models=(db.Feature,))
def task_formatter(view: ModelView, context: Any, model: db.Feature, value: List[str]) -> Markup:
    task_tracker_url = getattr(view, "task_tracker_url")
    if not task_tracker_url:
        return Markup(", ".join(value))
    task_links: List[str] = []
    for task in value:
        task_links.append(f"<a href='{task_tracker_url}/{task}' target='blank'>{task}</a>")
    return Markup(", ".join(task_links))


@safe_formatter(type=str, supported_models=(db.Feature,))
def file_path_formatter(view: ModelView, context: Any, model: db.Feature, value: str) -> Markup:
    if isinstance(model, db.Feature):
        value_without_suffix = re.sub(rf"({view.feature_suffix})", "", value)
        value_to_visualize = value_without_suffix.split("/")[-1]
        return Markup(f"<i>{value_to_visualize}</i>")
    raise NotImplementedError()


@safe_formatter(type=str, supported_models=(db.TestRun,))
def result_report_formatter(view: ModelView, context: Any, model: db.TestRun, value: str) -> Markup:
    report_status = TestReportStatus[getattr(model, "report_status")]
    test_run_id = getattr(model, "id")
    if test_run_id is None:
        raise ValueError("test_run_id could not be None!")
    report = getattr(model, "report")
    if report_status.has_report and isinstance(report, str):
        return Markup(
            f"<form action='{get_report_index_link(report)}' method='POST' target='_blank'>"
            f"<input type='hidden' name='run_id' value='{test_run_id}' />"
            f"<fieldset title='Go to report'>"
            f"<button class='link-button {get_button_class_by_status(value)}' type='submit'>{value}</button>"
            "</fieldset>"
            "</form>"
        )
    return Markup(
        f"<form action='{get_testrun_details_link(test_run_id)}'>"
        f"<fieldset title='Show details'>"
        f"<button class='link-button {get_button_class_by_status(value)}'>{value}</button>"
        "</fieldset>"
        "</form>"
    )


@safe_formatter(type=dict, supported_models=(db.TestUser,))
def json_formatter(view: ModelView, context: Any, model: db.BaseTable, value: Dict[str, Optional[str]]) -> Markup:
    info = ""
    for k, v in list(filter(lambda x: x, value.items())):
        info += f"<b>{k}</b>:&nbsp;&nbsp;{v}<br>"
    return Markup("<form>" "<fieldset>" f"<div class='json-data'>{info}</div>" "</fieldset>" "</form>")


@safe_formatter(type=str, supported_models=(db.Feature, db.TestRun))
def feature_link_formatter(view: ModelView, context: Any, model: db.BaseTable, value: str) -> Markup:
    if isinstance(model, db.Feature):
        return get_feature_link_markup(feature_id=model.id, feature_name=value)
    if isinstance(model, db.TestRun):
        return get_feature_link_markup(feature_id=model.scenario.feature_id, feature_name=value)
    raise NotImplementedError()


def _get_severity_color(severity: str) -> str:
    match severity:
        case allure.severity_level.BLOCKER:
            return "brown"
        case allure.severity_level.CRITICAL:
            return "maroon"
        case allure.severity_level.MINOR:
            return "gray"
        case allure.severity_level.TRIVIAL:
            return "silver"
    return "black"


@safe_formatter(type=str, supported_models=(db.Feature,))
def feature_severity_formatter(view: ModelView, context: Any, model: db.BaseTable, value: str) -> Markup:
    color = _get_severity_color(severity=value)
    return Markup(f"<font color='{color}'>{value.upper()}</font>")


@safe_formatter(type=int, supported_models=(db.Draft,))
def draft_feature_formatter(view: ModelView, context: Any, model: db.Draft, value: str) -> Markup:
    return get_feature_link_markup(feature_id=value, feature_name=model.feature.name)


@safe_formatter(type=int, supported_models=(db.Draft,))
def draft_testrun_formatter(view: ModelView, context: Any, model: db.BaseTable, value: int) -> Markup:
    return Markup(f"<a {get_testrun_details_link(value)}>{value}</a>")


@safe_formatter(type=str, supported_models=(db.Draft,))
def draft_prurl_formatter(view: ModelView, context: Any, model: db.BaseTable, value: str) -> Markup:
    return Markup(f"<a href='{URL(value).human_repr()}'>{value}</a>")



import logging
from pathlib import Path

from _pytest.nodes import Item
from pytest_bdd.parser import Scenario, ScenarioTemplate, Step

from overhave.entities import IFeatureExtractor
from overhave.pytest_plugin.helpers.allure_utils import add_git_project_feature_link_to_report
from overhave.storage import FeatureTypeName
from overhave.test_execution import OverhaveProjectSettings

logger = logging.getLogger(__name__)


def get_scenario(item: Item) -> Scenario:
    return item._obj.__scenario__  # type: ignore


def add_scenario_title_to_report(item: Item) -> None:
    item._obj.__allure_display_name__ = get_scenario(item).name  # type: ignore


def is_pytest_bdd_item(item: Item) -> bool:
    if hasattr(item, "_obj"):
        return hasattr(item._obj, "__scenario__") and isinstance(  # type: ignore
            get_scenario(item), (Scenario, ScenarioTemplate)
        )
    return False


def get_full_step_name(step: Step) -> str:
    return f"{step.keyword} {step._name}"


def set_git_project_url_if_necessary(
    project_settings: OverhaveProjectSettings,
    feature_extractor: IFeatureExtractor,
    item: Item,
    feature_type: FeatureTypeName,
) -> None:
    scenario = get_scenario(item)
    filename = Path(scenario.feature.filename)
    feature_type_dir = feature_extractor.feature_type_to_dir_mapping[feature_type]
    if not filename.is_relative_to(feature_type_dir):
        logger.warning("pytest_bdd item file '%s' is not relative to '%s'!")
        return
    relative_path = filename.relative_to(feature_type_dir.parent)
    add_git_project_feature_link_to_report(project_settings=project_settings, filepath=relative_path)



from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Type

from pydantic import BaseModel, validator
from yarl import URL

from overhave.base_settings import BaseOverhavePrefix
from overhave.utils import make_url


class EmptyOverhaveAdminURLError(ValueError):
    """Exception for situation with empty ````admin_url``` while trying to ```get_feature_url```."""


class EmptyTaskTrackerURLError(ValueError):
    """Exception for situation with empty ```task_tracker_url``` while trying to ```get_task_link```."""


class EmptyGitProjectURLError(ValueError):
    """Exception for situation with empty ```git_project_url``` while trying to ```get_git_feature_url```."""


class OverhaveAdminLinkSettings(BaseOverhavePrefix):
    """Settings for dynamic links to Overhave Admin in Allure report."""

    admin_url: Optional[URL]

    feature_id_filter_path: str = "feature/?flt2_0={feature_id}"
    feature_id_placeholder: str = "Overhave feature #{feature_id}"

    @validator("admin_url", pre=True)
    def make_admin_url(cls, v: Optional[str]) -> Optional[URL]:
        return make_url(v)

    @property
    def enabled(self) -> bool:
        return self.admin_url is not None

    def get_feature_url(self, feature_id: int) -> str:
        if isinstance(self.admin_url, URL):
            return self.admin_url.human_repr() + f"/{self.feature_id_filter_path.format(feature_id=feature_id)}"
        raise EmptyOverhaveAdminURLError("Overhave admin URL is None, so could not create link URL!")

    def get_feature_link_name(self, feature_id: int) -> str:
        return self.feature_id_placeholder.format(feature_id=feature_id)


class OverhaveProjectSettings(BaseOverhavePrefix):
    """
    Entity project settings.

    You could specify your fixture content for specific pytest usage with your own plugins.
    It could give you pretty extended file with powerful possibilities.
    `fixture_context` will be formatted by `OverhaveProccessor` before test would be started.
    """

    # Fixture content in list format, which would be compiled into formatted string.
    fixture_content: List[str] = [
        "from pytest_bdd import scenarios",
        "from overhave import overhave_proxy_manager",
        "pytest_plugins = overhave_proxy_manager().plugin_resolver.get_plugins()",
        "scenarios('{feature_file_path}')",
    ]

    # Task tracker URL
    task_tracker_url: Optional[URL]
    # Behaviour specification keyword for tasks attachment
    tasks_keyword: Optional[str]

    # Git project URL
    git_project_url: Optional[URL]

    # Templates are used for creation of test user and system specifications.
    # Keys for specification mapping are still the same - feature types.
    user_spec_template_mapping: Mapping[str, Type[BaseModel]] = {}

    @validator("task_tracker_url", pre=True)
    def make_tasktracker_url(cls, v: Optional[str]) -> Optional[URL]:
        return make_url(v)

    @validator("tasks_keyword")
    def validate_links_keyword(cls, v: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        if isinstance(v, str) and values.get("task_tracker_url") is None:
            raise ValueError("Task tracker URL should be specified in case of links keyword usage!")
        return v

    @validator("git_project_url", pre=True)
    def make_git_project_url(cls, v: Optional[str]) -> Optional[URL]:
        return make_url(v)

    def get_task_link(self, link: str) -> URL:
        if isinstance(self.task_tracker_url, URL):
            return self.task_tracker_url / link
        raise EmptyTaskTrackerURLError("Task tracker URL is None, so could not create link URL!")

    def get_git_feature_url(self, filepath: Path) -> URL:
        if isinstance(self.git_project_url, URL):
            return self.git_project_url / filepath.as_posix()
        raise EmptyGitProjectURLError("Git project URL is None, so could not create link URL!")


class OverhaveTestSettings(BaseOverhavePrefix):
    """Settings for PytestRunner, which runs scenario tests with specified addoptions."""

    default_pytest_addoptions: str = "--disable-warnings"
    extra_pytest_addoptions: Optional[str]

    workers: Optional[int]  # Number of xdist workers, `None` by default



from typing import List, Optional, cast

import allure
from pytest_bdd import types as default_types

from overhave.entities import OverhaveLanguageSettings, OverhaveScenarioCompilerSettings
from overhave.scenario.errors import IncorrectScenarioTextError
from overhave.scenario.mixin import PrefixMixin
from overhave.storage import TestExecutorContext


def generate_task_info(tasks: List[str], header: Optional[str]) -> str:
    if tasks and header is not None:
        return f"{header}: {', '.join(tasks)}"
    return ""


def generate_tags_list(context: TestExecutorContext) -> Optional[List[str]]:
    if feature_tags := [i.value for i in context.feature.feature_tags]:
        return feature_tags
    return None


class ScenarioCompiler(PrefixMixin):
    """Class for scenario compilation from text view into pytest_bdd feature format."""

    def __init__(
        self,
        compilation_settings: OverhaveScenarioCompilerSettings,
        language_settings: OverhaveLanguageSettings,
        tasks_keyword: Optional[str],
    ):
        self._compilation_settings = compilation_settings
        self._language_settings = language_settings
        self._tasks_keyword = tasks_keyword

    def _get_feature_type_tag(self, scenario_text: str, tag: str) -> str:
        if f"{self._compilation_settings.tag_prefix}{tag}" in scenario_text:
            return ""
        return f"{self._compilation_settings.tag_prefix}{tag}"

    def _get_additional_tags(self, scenario_text: str, tags: Optional[List[str]]) -> str:
        if f"{self._compilation_settings.tag_prefix}{tags}" in scenario_text:
            return ""
        if tags is not None:
            tags_with_prefix = (f"{self._compilation_settings.tag_prefix}{tag}" for tag in tags)
            return f"{' '.join(tags_with_prefix)}"
        return ""

    def _get_severity_tag(self, severity: allure.severity_level) -> str:
        return f"{self._compilation_settings.severity_prefix}{severity.value}"

    def _get_feature_prefix_if_specified(self, scenario_text: str) -> Optional[str]:
        keywords: List[str] = [default_types.FEATURE]
        if self._language_settings.step_prefixes is not None:
            keywords.append(self._language_settings.step_prefixes.FEATURE)

        for key in keywords:
            if self._as_prefix(key) not in scenario_text:
                continue
            return key
        return None

    def _detect_feature_prefix_by_scenario_format(self, scenario_text: str) -> str:
        if (
            self._as_prefix(default_types.SCENARIO) in scenario_text
            or self._as_prefix(default_types.SCENARIO_OUTLINE) in scenario_text
        ):
            return cast(str, default_types.FEATURE)
        step_prefixes = self._language_settings.step_prefixes
        if step_prefixes is not None and (
            self._as_prefix(step_prefixes.SCENARIO) in scenario_text
            or self._as_prefix(step_prefixes.SCENARIO_OUTLINE) in scenario_text
        ):
            return step_prefixes.FEATURE
        raise IncorrectScenarioTextError(
            "Could not find any scenario prefix in scenario text, so could not compile feature header!"
        )

    def _compile_header(self, context: TestExecutorContext) -> str:
        text = context.scenario.text
        feature_prefix = self._get_feature_prefix_if_specified(scenario_text=text)
        if not feature_prefix:
            feature_prefix = self._detect_feature_prefix_by_scenario_format(scenario_text=text)
        blocks_delimiter = f" {self._compilation_settings.blocks_delimiter} "
        if context.test_run.start is None:
            raise RuntimeError
        return "\n".join(
            (
                f"{self._get_feature_type_tag(scenario_text=text, tag=context.feature.feature_type.name)} "
                f"{self._get_additional_tags(scenario_text=text, tags=generate_tags_list(context))} "
                f"{self._get_severity_tag(severity=context.feature.severity)}",
                f"{self._as_prefix(feature_prefix)} {context.feature.name}",
                f"{self._compilation_settings.id_prefix} {context.feature.id}",
                (
                    f"{self._compilation_settings.created_by_prefix} {context.feature.author}"
                    f"{blocks_delimiter}"
                    f"{self._compilation_settings.last_edited_by_prefix} {context.feature.last_edited_by}"
                    f"{self._compilation_settings.time_delimiter} "
                    f"{context.feature.last_edited_at.strftime(self._compilation_settings.time_format)}"
                    f"{blocks_delimiter}"
                    f"{self._compilation_settings.published_by_prefix} {context.test_run.executed_by}"
                ),
                generate_task_info(tasks=context.feature.task, header=self._tasks_keyword),
                "",
            )
        )

    def compile(self, context: TestExecutorContext) -> str:
        return self._compile_header(context=context) + "\n" + context.scenario.text.strip("\n") + "\n"



import logging
from http import HTTPStatus
from typing import List, Optional

import fastapi

from overhave.api.deps import get_feature_type_storage, get_test_user_storage
from overhave.storage import (
    FeatureTypeName,
    FeatureTypeNotExistsError,
    IFeatureTypeStorage,
    ITestUserStorage,
    TestUserDoesNotExistError,
    TestUserModel,
    TestUserSpecification,
    TestUserUpdatingNotAllowedError,
)

logger = logging.getLogger(__name__)


def _get_test_user_by_id_handler(user_id: int, test_user_storage: ITestUserStorage) -> TestUserModel:
    logger.info("Getting %s with user_id=%s...", TestUserModel.__name__, user_id)
    test_user = test_user_storage.get_test_user_by_id(user_id)
    if test_user is None:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"User with id={user_id} does not exist!"
        )
    return test_user


def _get_test_user_by_name_handler(user_name: str, test_user_storage: ITestUserStorage) -> TestUserModel:
    logger.info("Getting %s with user_name='%s'...", TestUserModel.__name__, user_name)
    test_user = test_user_storage.get_test_user_by_name(user_name)
    if test_user is None:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"User with name='{user_name}' does not exist!"
        )
    return test_user


def get_test_user_handler(
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    test_user_storage: ITestUserStorage = fastapi.Depends(get_test_user_storage),
) -> TestUserModel:
    if user_id is not None:
        return _get_test_user_by_id_handler(user_id=user_id, test_user_storage=test_user_storage)
    if user_name is not None:
        return _get_test_user_by_name_handler(user_name=user_name, test_user_storage=test_user_storage)
    raise fastapi.HTTPException(
        status_code=HTTPStatus.BAD_REQUEST, detail="'user_id' or 'user_name' query parameter should be set!"
    )


def delete_test_user_handler(
    user_id: int,
    test_user_storage: ITestUserStorage = fastapi.Depends(get_test_user_storage),
) -> None:
    try:
        test_user_storage.delete_test_user(user_id)
    except TestUserDoesNotExistError:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"User with id={user_id} does not exist!"
        )


def test_user_list_handler(
    feature_type: FeatureTypeName,
    allow_update: bool,
    feature_type_storage: IFeatureTypeStorage = fastapi.Depends(get_feature_type_storage),
    test_user_storage: ITestUserStorage = fastapi.Depends(get_test_user_storage),
) -> List[TestUserModel]:
    logger.info(
        "Getting %s list with feature_type=%s and allow_update=%s ...",
        TestUserModel.__name__,
        feature_type,
        allow_update,
    )
    try:
        feature_type_model = feature_type_storage.get_feature_type_by_name(feature_type)
        return test_user_storage.get_test_users(feature_type_id=feature_type_model.id, allow_update=allow_update)
    except FeatureTypeNotExistsError:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"FeatureType with name='{feature_type}' does not exist!"
        )


def test_user_get_spec_handler(
    user_id: int, test_user_storage: ITestUserStorage = fastapi.Depends(get_test_user_storage)
) -> TestUserSpecification:
    test_user = _get_test_user_by_id_handler(user_id=user_id, test_user_storage=test_user_storage)
    return test_user.specification


def test_user_put_spec_handler(
    user_id: int,
    specification: TestUserSpecification,
    test_user_storage: ITestUserStorage = fastapi.Depends(get_test_user_storage),
) -> None:
    logger.info("Updating %s for user_id=%s...", TestUserSpecification.__name__, user_id)
    try:
        test_user_storage.update_test_user_specification(user_id=user_id, specification=specification)
    except TestUserDoesNotExistError:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"User with id={user_id} does not exist!"
        )
    except TestUserUpdatingNotAllowedError:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.BAD_REQUEST, detail=f"User's updating with id={user_id} not allowed!"
        )
    logger.info("%s for user_id=%s was successfully updated", TestUserSpecification.__name__, user_id)



import logging
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

from overhave.entities import IFeatureExtractor, OverhaveFileSettings
from overhave.scenario.compiler import ScenarioCompiler
from overhave.storage import TestExecutorContext
from overhave.test_execution.settings import OverhaveProjectSettings

logger = logging.getLogger(__name__)


class FileSavingMixin:
    """Mixin for files saving."""

    @staticmethod
    def _write_data(file: tempfile._TemporaryFileWrapper, data: str, entity_name: str) -> None:  # type: ignore
        file.write(data)
        file.flush()
        logger.debug("Generated %s file: '%s'", entity_name, file.name)
        assert Path(file.name).exists(), f"{entity_name.upper()} file not created!"


class FileManager(FileSavingMixin):
    """Class for feature files management."""

    def __init__(
        self,
        project_settings: OverhaveProjectSettings,
        file_settings: OverhaveFileSettings,
        feature_extractor: IFeatureExtractor,
        scenario_compiler: ScenarioCompiler,
    ):
        self._project_settings = project_settings
        self._file_settings = file_settings
        self._feature_extractor = feature_extractor
        self._scenario_compiler = scenario_compiler

    @contextmanager
    def tmp_feature_file(
        self, context: TestExecutorContext
    ) -> Iterator[tempfile._TemporaryFileWrapper]:  # type: ignore
        file_name = Path(context.feature.file_path).name
        logger.debug("Feature file name: '%s'", file_name)
        with tempfile.NamedTemporaryFile(
            dir=self._file_settings.tmp_features_dir,
            prefix=f"{file_name}_id{context.feature.id}",
            suffix=self._file_settings.feature_suffix,
            mode="w",
        ) as file:
            data = self._scenario_compiler.compile(context=context)
            logger.debug("Scenario file:\n%s", data)
            self._write_data(file=file, data=data, entity_name="feature")
            yield file

    @contextmanager
    def tmp_fixture_file(
        self, context: TestExecutorContext, feature_file: tempfile._TemporaryFileWrapper  # type: ignore
    ) -> Iterator[tempfile._TemporaryFileWrapper]:  # type: ignore
        with tempfile.NamedTemporaryFile(
            dir=self._file_settings.tmp_fixtures_dir,
            prefix=f"{context.test_run.id}_",
            suffix=self._file_settings.fixture_suffix,
            mode="w",
        ) as file:
            joined_content = "\n".join(self._project_settings.fixture_content)
            data = joined_content.format(feature_file_path=feature_file.name)

            logger.debug("Fixture file:\n%s", data)
            self._write_data(file=file, data=data, entity_name="fixture")
            yield file

    def produce_feature_file(self, context: TestExecutorContext) -> Path:
        feature_file_path = (
            self._feature_extractor.feature_type_to_dir_mapping[context.feature.feature_type.name]
            / context.feature.file_path
        )
        logger.info("Write scenario to feature file '%s'...", feature_file_path.as_posix())

        scenario_text = self._scenario_compiler.compile(context=context)
        logger.debug(scenario_text)
        feature_file_path.parent.mkdir(parents=True, exist_ok=True)
        feature_file_path.write_text(scenario_text)

        logger.info("Scenario has successfully written")
        return feature_file_path



import logging
from http import HTTPStatus
from typing import List, Optional

import fastapi

from overhave.api.deps import get_feature_storage, get_feature_tag_storage
from overhave.api.views.tags_views import tags_item_handler
from overhave.storage import FeatureModel, IFeatureStorage, IFeatureTagStorage

logger = logging.getLogger(__name__)


def get_features_handler(
    tag_id: Optional[int] = None,
    tag_value: Optional[str] = None,
    feature_storage: IFeatureStorage = fastapi.Depends(get_feature_storage),
    tag_storage: IFeatureTagStorage = fastapi.Depends(get_feature_tag_storage),
) -> List[FeatureModel]:
    if tag_id is not None:
        logger.info("Getting %s by tag_id=%s...", FeatureModel.__name__, tag_id)
        return feature_storage.get_features_by_tag(tag_id=tag_id)
    if tag_value is not None:
        logger.info("Getting %s by tag_value=%s...", FeatureModel.__name__, tag_value)
        tag_model = tags_item_handler(value=tag_value, feature_tag_storage=tag_storage)
        return feature_storage.get_features_by_tag(tag_id=tag_model.id)
    raise fastapi.HTTPException(
        status_code=HTTPStatus.BAD_REQUEST, detail="'tag_id' or 'tag_value' query parameter should be set"
    )
